"""
Credits:

* `gist@ninely <https://gist.github.com/ninely/88485b2e265d852d3feb8bd115065b1a>`_
* `langchain@#1705 <https://github.com/hwchase17/langchain/discussions/1706>`_
"""
import asyncio
import logging
from functools import partial, wraps
from typing import Any, Awaitable, Callable, Optional, Union

import aiohttp
from fastapi.responses import StreamingResponse as _StreamingResponse
from langchain.chains.base import Chain
from starlette.background import BackgroundTask
from starlette.types import Receive, Scope, Send

from lanarky.callbacks import (
    AsyncLanarkyCallback,
    get_streaming_callback,
    get_streaming_json_callback,
)

logger = logging.getLogger(__name__)


def openai_aiosession(func):
    """Decorator to set openai.aiosession for StreamingResponse."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            import openai  # type: ignore
        except ImportError:
            raise ImportError(
                "openai is not installed. Install it with `pip install 'lanarky[openai]'`."
            )

        openai.aiosession.set(aiohttp.ClientSession())
        logger.debug(f"opeanai.aiosession set: {openai.aiosession.get()}")

        try:
            await func(*args, **kwargs)
        finally:
            await openai.aiosession.get().close()
            logger.debug(f"opeanai.aiosession closed: {openai.aiosession.get()}")

    return wrapper


# TODO: create OpenAIStreamingResponse for streaming with OpenAI only
class StreamingResponse(_StreamingResponse):
    """StreamingResponse class wrapper for langchain chains."""

    def __init__(
        self,
        chain_executor: Callable[[Send], Awaitable[Any]],
        background: Optional[BackgroundTask] = None,
        **kwargs: Any,
    ) -> None:
        """Constructor method.

        Args:
            chain_executor: function to execute ``chain.acall()``.
            background: A ``BackgroundTask`` object to run in the background.
        """
        super().__init__(content=iter(()), background=background, **kwargs)

        self.chain_executor = chain_executor

    async def listen_for_disconnect(self, receive: Receive) -> None:
        """Listen for client disconnect."""
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                logger.debug("Client disconnected")
                break

    async def stream_response(self, send: Send) -> None:
        """Streams the response."""
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )

        try:
            outputs = await self.chain_executor(send)
            if self.background is not None:
                self.background.kwargs["outputs"] = outputs
        except Exception as e:
            if self.background is not None:
                self.background.kwargs["outputs"] = str(e)
            await send(
                {
                    "type": "http.response.body",
                    "body": str(e).encode(self.charset),
                    "more_body": False,
                }
            )
            return

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    @openai_aiosession
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async def wrap(func: Callable[[], Awaitable[None]]) -> None:
            await func()
            raise asyncio.CancelledError

        async def run_tasks():
            stream_response_task = asyncio.create_task(
                wrap(partial(self.stream_response, send))
            )
            listen_for_disconnect_task = asyncio.create_task(
                wrap(partial(self.listen_for_disconnect, receive))
            )

            try:
                await asyncio.gather(stream_response_task, listen_for_disconnect_task)
            except asyncio.CancelledError:
                pass

        await asyncio.create_task(run_tasks())

        if self.background is not None:
            await self.background()

    @staticmethod
    def _create_chain_executor(
        chain: Chain,
        inputs: Union[dict[str, Any], Any],
        as_json: bool = False,
        callback: Optional[AsyncLanarkyCallback] = None,
        **callback_kwargs,
    ) -> Callable[[Send], Awaitable[Any]]:
        """Creates a function to execute ``chain.acall()``.

        Args:
            chain: A ``Chain`` object.
            inputs: Inputs to pass to ``chain.acall()``.
            as_json: Whether to return the outputs as JSON.
            callback_kwargs: Keyword arguments to pass to the callback function.
        """
        if callback is None:
            get_callback_fn = (
                get_streaming_json_callback if as_json else get_streaming_callback
            )
            callback = partial(get_callback_fn, chain)

        async def wrapper(send: Send):
            return await chain.acall(
                inputs=inputs,
                callbacks=[callback(send=send, **callback_kwargs)],
            )

        return wrapper

    @classmethod
    def from_chain(
        cls,
        chain: Chain,
        inputs: Union[dict[str, Any], Any],
        as_json: bool = False,
        background: Optional[BackgroundTask] = None,
        callback: Optional[AsyncLanarkyCallback] = None,
        callback_kwargs: dict[str, Any] = {},
        **kwargs: Any,
    ) -> "StreamingResponse":
        """Creates a ``StreamingResponse`` object from a ``Chain`` object.

        Args:
            chain: A ``Chain`` object.
            inputs: Inputs to pass to ``chain.acall()``.
            as_json: Whether to return the outputs as JSON.
            background: A ``BackgroundTask`` object to run in the background.
            callback: custom callback function to use instead of using the registry.
            callback_kwargs: Keyword arguments to pass to the callback function.
        """
        chain_executor = cls._create_chain_executor(
            chain, inputs, as_json, callback, **callback_kwargs
        )

        return cls(
            chain_executor=chain_executor,
            background=background,
            **kwargs,
        )



from typing import Any

from pydantic import BaseModel


class StreamingJSONResponse(BaseModel):
    """Streaming JSON response."""

    token: str = ""


class AnswerStreamingJSONResponse(BaseModel):
    """Answer response used when cache is enabled and tokens haven't been streamed.
    Should only be output when on_llm_new_token hasn't been invoked before on_chain_end.
    """

    answer: str = ""  # only returned when langchain.llm_cache is used


class BaseRetrievalQAStreamingJSONResponse(BaseModel):
    """Base class for retrieval-based QA streaming JSON response."""

    source_documents: list[dict[str, Any]]



from enum import IntEnum
from functools import lru_cache
from typing import Type

from fastapi import Depends, WebSocket, params
from langchain.chains.base import Chain
from langchain.schema import Document
from pydantic import BaseModel, create_model

from lanarky.responses import StreamingResponse
from lanarky.websockets import WebsocketConnection


class StreamingMode(IntEnum):
    """Streaming modes for LangchainRouter."""

    OFF = 0
    TEXT = 1
    JSON = 2


class LLMCacheMode(IntEnum):
    """LLM cache modes for LangchainRouter."""

    OFF = 0
    IN_MEMORY = 1
    REDIS = 2
    GPTCACHE = 3


BASE_LANGCHAIN_TYPES = [
    "LLMChain",
    "ConversationChain",
    "AgentExecutor",
    "RetrievalQAWithSourcesChain",
    "ConversationalRetrievalChain",
]
ERROR_MESSAGE = """Error! Creating a {model_type} model for '{chain_type}' is not currently supported by 'LangchainRouter.add_langchain_api_route()'.
Available chain types: {chain_types}

To use a custom chain type, you must define your own FastAPI endpoint.
"""
CHAT_HISTORY_KEY = "chat_history"
SOURCE_DOCUMENTS_KEY = "source_documents"


def create_langchain_dependency(langchain_object: Type[Chain]) -> params.Depends:
    """Creates a langchain object dependency.

    Args:
        langchain_object: The langchain object.
    """

    @lru_cache(maxsize=1)
    def dependency() -> Chain:
        return langchain_object

    return Depends(dependency)


def create_request_from_langchain_dependency(
    langchain_dependency: params.Depends, name_prefix: str = ""
) -> Type[BaseModel]:
    """Creates a request model from a langchain dependency.

    Args:
        langchain_dependency: The langchain dependency.
        name_prefix: The name prefix for the model.
    """
    langchain_object: Chain = langchain_dependency.dependency()
    langchain_object_name = str(langchain_object.__class__.__name__)
    model_name = f"{name_prefix}{langchain_object_name}Request"

    additional_keys = (
        {CHAT_HISTORY_KEY: (list[tuple[str, str]], ...)}
        if langchain_object_name == "ConversationalRetrievalChain"
        else {}
    )

    if langchain_object_name in BASE_LANGCHAIN_TYPES:
        return create_model(
            model_name,
            **{
                **{key: (str, ...) for key in langchain_object.input_keys},
                **additional_keys,
            },
        )
    else:
        raise TypeError(
            ERROR_MESSAGE.format(
                model_type="Request",
                chain_type=langchain_object_name,
                chain_types=BASE_LANGCHAIN_TYPES,
            )
        )


def create_response_model_from_langchain_dependency(
    langchain_dependency: params.Depends, name_prefix: str = ""
) -> Type[BaseModel]:
    """Creates a response model from a langchain dependency.

    Args:
        langchain_dependency: The langchain dependency.
        name_prefix: The name prefix for the model.
    """
    langchain_object: Chain = langchain_dependency.dependency()
    langchain_object_name = str(langchain_object.__class__.__name__)
    model_name = f"{name_prefix}{langchain_object_name}Response"

    additional_keys = (
        {SOURCE_DOCUMENTS_KEY: (list[Document], ...)}
        if hasattr(langchain_object, "return_source_documents")
        and langchain_object.return_source_documents
        else {}
    )

    if langchain_object_name in BASE_LANGCHAIN_TYPES:
        return create_model(
            model_name,
            **{
                **{key: (str, ...) for key in langchain_object.output_keys},
                **additional_keys,
            },
        )
    else:
        raise TypeError(
            ERROR_MESSAGE.format(
                model_type="Response",
                chain_type=langchain_object_name,
                chain_types=BASE_LANGCHAIN_TYPES,
            )
        )


def create_langchain_base_endpoint(
    endpoint_request: BaseModel,
    langchain_dependency: params.Depends,
    response_model: BaseModel,
):
    """Creates a base Langchain endpoint.

    Args:
        endpoint_request: The request model.
        langchain_dependency: The langchain dependency.
        response_model: The response model.
    """

    async def endpoint(
        request: endpoint_request,
        langchain_object: Chain = langchain_dependency,
    ) -> response_model:
        """Base chat endpoint."""
        return await langchain_object.acall(
            inputs=request.dict(), return_only_outputs=True
        )

    return endpoint


def create_langchain_streaming_endpoint(
    endpoint_request: BaseModel, langchain_dependency: params.Depends
):
    """Creates a streaming Langchain endpoint.

    Args:
        endpoint_request: The request model.
        langchain_dependency: The langchain dependency.
    """

    async def endpoint(
        request: endpoint_request,
        langchain_object: Chain = langchain_dependency,
    ) -> StreamingResponse:
        """Streaming chat endpoint."""
        return StreamingResponse.from_chain(
            langchain_object, request.dict(), media_type="text/event-stream"
        )

    return endpoint


def create_langchain_streaming_json_endpoint(
    endpoint_request: BaseModel, langchain_dependency: params.Depends
):
    """Creates a streaming JSON Langchain endpoint.

    Args:
        endpoint_request: The request model.
        langchain_dependency: The langchain dependency.
    """

    async def endpoint(
        request: endpoint_request,
        langchain_object: Chain = langchain_dependency,
    ) -> StreamingResponse:
        """Streaming JSON chat endpoint."""
        return StreamingResponse.from_chain(
            langchain_object,
            request.dict(),
            as_json=True,
            media_type="text/event-stream",
        )

    return endpoint


def create_langchain_endpoint(
    endpoint_request: BaseModel,
    langchain_dependency: params.Depends,
    response_model: BaseModel,
    streaming_mode: StreamingMode,
):
    """Creates a Langchain endpoint based on the streaming mode.

    Args:
        endpoint_request: The request model.
        langchain_dependency: The langchain dependency.
        response_model: The response model.
        streaming_mode: The streaming mode.
    """
    if streaming_mode == StreamingMode.OFF:
        endpoint = create_langchain_base_endpoint(
            endpoint_request, langchain_dependency, response_model
        )
    elif streaming_mode == StreamingMode.TEXT:
        endpoint = create_langchain_streaming_endpoint(
            endpoint_request, langchain_dependency
        )
    elif streaming_mode == StreamingMode.JSON:
        endpoint = create_langchain_streaming_json_endpoint(
            endpoint_request, langchain_dependency
        )
    else:
        raise ValueError(f"Invalid streaming mode: {streaming_mode}")

    return endpoint


def create_langchain_websocket_endpoint(
    websocket: WebSocket, langchain_dependency: params.Depends
):
    """Creates a websocket Langchain endpoint.

    Args:
        websocket: The websocket model.
        langchain_dependency: The langchain dependency.
    """

    async def endpoint(
        websocket: websocket,
        langchain_object: Chain = langchain_dependency,
    ) -> None:
        """Websocket chat endpoint."""
        connection = WebsocketConnection.from_chain(
            chain=langchain_object, websocket=websocket
        )
        await connection.connect()

    return endpoint



from typing import Any

import requests
from fastapi import FastAPI

from .settings import get_settings


# FIXME: This is a hacky way to get the query key from the OpenAPI spec
def get_payload_query_key() -> str:
    openapi_json_url = f"{get_settings().api_url}/openapi.json"
    response = requests.get(openapi_json_url)
    openapi_data = response.json()

    ref = openapi_data["paths"][get_settings().api_endpoint]["post"]["requestBody"][
        "content"
    ]["application/json"]["schema"]["$ref"]
    schema_name = ref.split("/")[-1]
    required = openapi_data["components"]["schemas"][schema_name]["required"]

    return required[0]


def send_query(api_url: str, query: str, chat: list[Any], history: list[Any]):
    """Adapted code from: https://huggingface.co/spaces/ysharma/Gradio-demo-streaming"""
    history = history or []
    history.append(query)

    query_key = get_payload_query_key()

    payload = {query_key: query}
    headers = {
        "accept": "text/event-stream",
        "Content-Type": "application/json",
        "Connection": "keep-alive",
    }

    try:
        response = requests.post(
            api_url, headers=headers, json=payload, timeout=60, stream=True
        )
        response.raise_for_status()  # Raise stored HTTPError, if one occurred

    except requests.exceptions.HTTPError as httpErr:
        print(f"HTTP error occurred: {httpErr}")
        return
    except requests.exceptions.ConnectionError as connErr:
        print(f"Error connecting: {connErr}")
        return
    except requests.exceptions.Timeout as timeOutErr:
        print(f"Timeout error: {timeOutErr}")
        return
    except requests.exceptions.RequestException as reqErr:
        print(f"Something went wrong with the request: {reqErr}")
        return

    token_counter = 0
    partial_words = ""
    for chunk in response.iter_content(chunk_size=1024):  # Smaller chunk size
        if chunk:
            partial_words = partial_words + chunk.decode()
            if token_counter == 0:
                history.append(" " + partial_words)
            else:
                history[-1] = partial_words
            chat = [(history[i], history[i + 1]) for i in range(0, len(history) - 1, 2)]
            token_counter += 1
            yield chat, history


def clear_chat():
    return [], []


def mount_gradio_app(
    app: FastAPI,
    path: str = get_settings().gradio_path,
    chat_endpoint: str = get_settings().api_endpoint,
    title: str = get_settings().title,
):
    """Mounts a Gradio app on a FastAPI app."""

    try:
        import gradio as gr
    except ImportError:
        raise ImportError(
            "Please install gradio to use this feature: pip install gradio"
        )

    blocks = gr.Blocks(title=title)

    with blocks:
        gr.Markdown(f"<h3><center>{title}</center></h3>")

        with gr.Row():
            query = gr.Textbox(
                label="Enter your question",
                lines=1,
            )
            submit = gr.Button(value="Ask", variant="secondary").style(full_width=False)
            clear = gr.Button(value="Clear", variant="secondary").style(
                full_width=False
            )

        api_url = gr.State(get_settings().api_url + chat_endpoint)
        chatbot = gr.Chatbot()
        history = gr.State()

        submit.click(
            send_query,
            inputs=[api_url, query, chatbot, history],
            outputs=[chatbot, history],
        )
        query.submit(
            send_query,
            inputs=[api_url, query, chatbot, history],
            outputs=[chatbot, history],
        )
        clear.click(
            clear_chat,
            outputs=[chatbot, history],
        )

    return gr.mount_gradio_app(app=app, blocks=blocks.queue(), path=path)



"""
Credits:

* `chat-langchain <https://github.com/hwchase17/chat-langchain>`_
* `langchain-chat-websockets <https://github.com/pors/langchain-chat-websockets>`_
"""
import logging
from functools import partial
from typing import Any, Awaitable, Callable, Optional

from fastapi import WebSocket, WebSocketDisconnect
from langchain.chains.base import Chain
from pydantic import BaseModel, Field

from lanarky.callbacks import AsyncLanarkyCallback, get_websocket_callback
from lanarky.schemas import Message, MessageType, Sender, WebsocketResponse

logger = logging.getLogger(__name__)


class BaseWebsocketConnection(BaseModel):
    """Base class for websocket connections."""

    websocket: WebSocket = Field(...)
    chain_executor: Callable[[str], Awaitable[Any]] = Field(...)
    connection_accepted: bool = Field(False)

    class Config:
        arbitrary_types_allowed = True

    async def connect(self, accept_connection: bool = True):
        if accept_connection and self.connection_accepted:
            raise RuntimeError("Connection already accepted.")

        if accept_connection:
            """Connect to websocket."""
            await self.websocket.accept()
            self.connection_accepted = True

        while True:
            try:
                user_message = await self.websocket.receive_text()
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.HUMAN,
                        message=user_message,
                        message_type=MessageType.STREAM,
                    ).dict()
                )
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.BOT,
                        message=Message.NULL,
                        message_type=MessageType.START,
                    ).dict()
                )
                await self.chain_executor(user_message)
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.BOT,
                        message=Message.NULL,
                        message_type=MessageType.END,
                    ).dict()
                )
            except WebSocketDisconnect:
                logger.debug("client disconnected.")
                break
            except Exception as e:
                logger.error(e)
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.BOT,
                        message=Message.ERROR,
                        message_type=MessageType.ERROR,
                    ).dict()
                )

    @staticmethod
    def _create_chain_executor(
        chain: Chain,
        websocket: WebSocket,
        response: WebsocketResponse,
        callback: Optional[AsyncLanarkyCallback] = None,
        **callback_kwargs,
    ) -> Callable[[str], Awaitable[Any]]:
        """Creates a function to execute ``chain.acall()``.

        Args:
            chain: langchain chain instance.
            websocket: websocket instance.
            response: WebsocketResponse instance.
            callback: custom callback function.
            callback_kwargs: keyword arguments for callback function.
        """
        raise NotImplementedError

    @classmethod
    def from_chain(
        cls,
        chain: Chain,
        websocket: WebSocket,
        callback: Optional[AsyncLanarkyCallback] = None,
        callback_kwargs: dict[str, Any] = {},
    ) -> "BaseWebsocketConnection":
        """Creates a BaseWebsocketConnection instance from a langchain chain instance.

        Args:
            chain: langchain chain instance.
            websocket: websocket instance.
            callback: custom callback function.
            callback_kwargs: keyword arguments for callback function.
        """
        chain_executor = cls._create_chain_executor(
            chain,
            websocket,
            WebsocketResponse(
                sender=Sender.BOT, message=Message.NULL, message_type=MessageType.STREAM
            ),
            callback=callback,
            **callback_kwargs,
        )

        return cls(
            chain_executor=chain_executor,
            websocket=websocket,
        )


class WebsocketConnection(BaseWebsocketConnection):
    """BaseLangchainStreamingResponse class wrapper for LLMChain instances."""

    @staticmethod
    def _create_chain_executor(
        chain: Chain,
        websocket: WebSocket,
        response: WebsocketResponse,
        callback: Optional[AsyncLanarkyCallback] = None,
        **callback_kwargs,
    ) -> Callable[[str], Awaitable[Any]]:
        """Creates a function to execute ``chain.acall()``.

        Args:
            chain: langchain chain instance.
            websocket: websocket instance.
            response: WebsocketResponse instance.
            callback_kwargs: keyword arguments for callback function.
        """
        if callback is None:
            callback = partial(get_websocket_callback, chain)

        async def wrapper(user_message: str):
            return await chain.acall(
                inputs=user_message,
                callbacks=[
                    callback(
                        websocket=websocket,
                        response=response,
                        **callback_kwargs,
                    )
                ],
            )

        return wrapper



from typing import Any, Callable, Union

from .base import register

STREAMING_CALLBACKS: dict[str, Any] = {}
WEBSOCKET_CALLBACKS: dict[str, Any] = {}
STREAMING_JSON_CALLBACKS: dict[str, Any] = {}


def register_streaming_callback(key: Union[list[str], str], **kwargs) -> Callable:
    """Register a streaming callback handler."""

    def _register_cls(cls: Any) -> Callable:
        register(key, STREAMING_CALLBACKS, **kwargs)(cls=cls)
        return cls

    return _register_cls


def register_websocket_callback(key: Union[list[str], str], **kwargs) -> Callable:
    """Register a websocket callback handler."""

    def _register_cls(cls: Any) -> Callable:
        register(key, WEBSOCKET_CALLBACKS, **kwargs)(cls=cls)
        return cls

    return _register_cls


def register_streaming_json_callback(key: Union[list[str], str], **kwargs) -> Callable:
    """Register an streaming json callback handler."""

    def _register_cls(cls: Any) -> Callable:
        register(key, STREAMING_JSON_CALLBACKS, **kwargs)(cls=cls)
        return cls

    return _register_cls



from .callbacks import (
    AnswerStreamingJSONResponse,
    BaseRetrievalQAStreamingJSONResponse,
    StreamingJSONResponse,
)
from .websockets import Message, MessageType, Sender, WebsocketResponse

__all__ = [
    "AnswerStreamingJSONResponse",
    "StreamingJSONResponse",
    "BaseRetrievalQAStreamingJSONResponse",
    "WebsocketResponse",
    "Sender",
    "Message",
    "MessageType",
]



from .callbacks import (
    STREAMING_CALLBACKS,
    STREAMING_JSON_CALLBACKS,
    WEBSOCKET_CALLBACKS,
    register_streaming_callback,
    register_streaming_json_callback,
    register_websocket_callback,
)

__all__ = [
    "STREAMING_CALLBACKS",
    "STREAMING_JSON_CALLBACKS",
    "WEBSOCKET_CALLBACKS",
    "register_streaming_callback",
    "register_streaming_json_callback",
    "register_websocket_callback",
]



import json
from abc import abstractmethod
from typing import Any

from fastapi import WebSocket
from langchain.callbacks.base import AsyncCallbackHandler
from langchain.globals import get_llm_cache
from pydantic import BaseModel, Field
from starlette.types import Message, Send

from lanarky.schemas import StreamingJSONResponse, WebsocketResponse


class AsyncLanarkyCallback(AsyncCallbackHandler, BaseModel):
    """Async Callback handler for FastAPI StreamingResponse."""

    output_key: str = Field(default="answer")

    llm_cache_used: bool = Field(default_factory=lambda: get_llm_cache() is not None)

    @property
    def llm_cache_enabled(self) -> bool:
        """Determine if LLM caching is enabled."""
        return get_llm_cache() is not None

    @property
    def always_verbose(self) -> bool:
        """Whether to call verbose callbacks even if verbose is False."""
        return True

    class Config:
        arbitrary_types_allowed = True

    @abstractmethod
    def _construct_message(self, content: Any) -> Any:  # pragma: no cover
        """Constructs a Message from a string."""
        pass


class AsyncStreamingResponseCallback(AsyncLanarkyCallback):
    """Async Callback handler for StreamingResponse."""

    send: Send = Field(...)

    def _construct_message(self, content: str) -> Message:
        """Constructs a Message from a string."""
        return {
            "type": "http.response.body",
            "body": content.encode("utf-8"),
            "more_body": True,
        }


class AsyncWebsocketCallback(AsyncLanarkyCallback):
    """Async Callback handler for WebsocketConnection."""

    websocket: WebSocket = Field(...)
    response: WebsocketResponse = Field(...)

    def _construct_message(self, content: str) -> dict:
        """Constructs a WebsocketResponse from a string."""
        return {**self.response.dict(), **{"message": content}}


class AsyncStreamingJSONResponseCallback(AsyncStreamingResponseCallback):
    """Async Callback handler for StreamingJSONResponse."""

    send: Send = Field(...)

    def _construct_message(self, content: StreamingJSONResponse) -> Message:
        """Constructs a Message from a dictionary."""
        return {
            "type": "http.response.body",
            "body": json.dumps(
                content.dict(),
                ensure_ascii=False,
                allow_nan=False,
                indent=None,
                separators=(",", ":"),
            ).encode("utf-8"),
            "more_body": True,
        }



from typing import Any

from lanarky.register import (
    register_streaming_callback,
    register_streaming_json_callback,
    register_websocket_callback,
)
from lanarky.schemas import StreamingJSONResponse

from .base import AsyncLanarkyCallback
from .llm import (
    AsyncLLMChainStreamingCallback,
    AsyncLLMChainStreamingJSONCallback,
    AsyncLLMChainWebsocketCallback,
)


class AsyncAgentsLanarkyCallback(AsyncLanarkyCallback):
    """Base AsyncCallback handler for AgentExecutor.

    Adapted from `langchain/callbacks/streaming_stdout_final_only.py <https://github.com/hwchase17/langchain/blob/master/langchain/callbacks/streaming_stdout_final_only.py>`_
    """

    answer_prefix_tokens: list[str] = ["Final", " Answer", ":"]
    last_tokens: list[str] = [""] * len(answer_prefix_tokens)
    answer_reached: bool = False

    async def on_llm_start(
        self, serialized: dict[str, Any], prompts: list[str], **kwargs: Any
    ) -> None:
        """Run when LLM starts running."""
        self.last_tokens = [""] * len(self.answer_prefix_tokens)
        self.answer_reached = False

    def _check_if_answer_reached(self, token: str):
        """Checks if the final answer has been reached."""
        self.last_tokens.append(token)
        if len(self.last_tokens) > len(self.answer_prefix_tokens):
            self.last_tokens.pop(0)

        if self.last_tokens == self.answer_prefix_tokens:
            self.answer_reached = True
            return

        return self.answer_reached


@register_streaming_callback("AgentExecutor")
class AsyncAgentsStreamingCallback(
    AsyncAgentsLanarkyCallback, AsyncLLMChainStreamingCallback
):
    """AsyncStreamingResponseCallback handler for AgentExecutor."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        if self._check_if_answer_reached(token):
            message = self._construct_message(token)
            await self.send(message)


@register_websocket_callback("AgentExecutor")
class AsyncAgentsWebsocketCallback(
    AsyncAgentsLanarkyCallback, AsyncLLMChainWebsocketCallback
):
    """AsyncWebsocketCallback handler for AgentExecutor."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        if self._check_if_answer_reached(token):
            message = self._construct_message(token)
            await self.websocket.send_json(message)


@register_streaming_json_callback("AgentExecutor")
class AsyncAgentsStreamingJSONCallback(
    AsyncAgentsLanarkyCallback, AsyncLLMChainStreamingJSONCallback
):
    """AsyncStreamingJSONCallback handler for AgentExecutor."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        if self._check_if_answer_reached(token):
            message = self._construct_message(StreamingJSONResponse(token=token))
            await self.send(message)



from enum import Enum
from typing import Union

from pydantic import BaseModel


class Sender(str, Enum):
    """Sender of a websocket message."""

    BOT = "bot"
    HUMAN = "human"


class Message(str, Enum):
    """Message types for websocket messages."""

    NULL = ""
    ERROR = "Sorry, something went wrong. Try again."


class MessageType(str, Enum):
    """Message types for websocket messages."""

    START = "start"
    STREAM = "stream"
    END = "end"
    ERROR = "error"
    INFO = "info"


class WebsocketResponse(BaseModel):
    """Websocket response."""

    sender: Sender
    message: Union[Message, str]
    message_type: MessageType

    class Config:
        use_enum_values = True



from typing import Any, Optional, Union


def register(
    key: Union[list[str], str],
    _registry: dict[str, tuple[Any, list[str]]],
    *,
    override: bool = False,
) -> Any:
    """Add a class/function to a registry with required keyword arguments.

    ``_registry`` is a dictionary mapping from a key to a tuple of the class/function
    and a list of required keyword arguments, if keyword arguments are passed. Otherwise
    it is a dictionary mapping from a key to the class/function.

    Args:
        key: key or list of keys to register the class/function under.
        _registry: registry to add the class/function to.
        override: if True, override existing keys in the registry.
    """

    def _register_cls(cls: Any, required_kwargs: Optional[list] = None) -> Any:
        if isinstance(key, str):
            keys = [key]
        else:
            keys = key

        for _key in keys:
            if _key in _registry and not override:
                raise KeyError(f"{cls} already registered as {_key}")
            _registry[_key] = cls if required_kwargs is None else (cls, required_kwargs)
        return cls

    return _register_cls



from typing import Any, Optional, Type

from fastapi.routing import APIRouter
from fastapi.websockets import WebSocket
from langchain.chains.base import Chain

from .utils import (
    LLMCacheMode,
    StreamingMode,
    create_langchain_dependency,
    create_langchain_endpoint,
    create_langchain_websocket_endpoint,
    create_request_from_langchain_dependency,
    create_response_model_from_langchain_dependency,
)


class LangchainRouter(APIRouter):
    def __init__(
        self,
        *,
        langchain_url: Optional[str] = None,
        langchain_object: Optional[Type[Chain]] = None,
        langchain_endpoint_kwargs: Optional[dict[str, Any]] = None,
        streaming_mode: Optional[StreamingMode] = None,
        llm_cache_mode: Optional[LLMCacheMode] = None,
        llm_cache_kwargs: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.langchain_url = langchain_url
        self.langchain_object = langchain_object
        self.langchain_endpoint_kwargs = langchain_endpoint_kwargs or {}
        self.streaming_mode = streaming_mode
        self.llm_cache_mode = llm_cache_mode
        self.llm_cache_kwargs = llm_cache_kwargs or {}

        self.langchain_dependencies = []

        self.setup()

    def setup(self) -> None:
        """Sets up the Langchain router."""
        if self.langchain_url is not None:
            self.add_langchain_api_route(
                self.langchain_url,
                self.langchain_object,
                self.streaming_mode,
                **self.langchain_endpoint_kwargs,
            )

        if self.llm_cache_mode is not None:
            self.setup_llm_cache()

    def setup_llm_cache(self) -> None:
        """Sets up the LLM cache."""
        from langchain.globals import set_llm_cache

        if self.llm_cache_mode == LLMCacheMode.IN_MEMORY:
            from langchain.cache import InMemoryCache

            set_llm_cache(InMemoryCache())

        elif self.llm_cache_mode == LLMCacheMode.REDIS:
            try:
                from redis import Redis  # type: ignore
            except ImportError:
                raise ImportError(
                    """Redis is not installed. Install it with `pip install "lanarky[redis]"`."""
                )
            from langchain.cache import RedisCache

            set_llm_cache(RedisCache(redis_=Redis.from_url(**self.llm_cache_kwargs)))

        elif self.llm_cache_mode == LLMCacheMode.GPTCACHE:
            try:
                from gptcache import Cache  # type: ignore
                from gptcache.manager.factory import manager_factory  # type: ignore
                from gptcache.processor.pre import get_prompt  # type: ignore
            except ImportError:
                raise ImportError(
                    """GPTCache is not installed. Install it with `pip install "lanarky[gptcache]"`."""
                )
            import hashlib

            from langchain.cache import GPTCache

            def init_gptcache(cache_obj: Cache, llm: str):
                hashed_llm = hashlib.sha256(llm.encode()).hexdigest()
                cache_obj.init(
                    pre_embedding_func=get_prompt,
                    data_manager=manager_factory(
                        manager="map", data_dir=f"map_cache_{hashed_llm}"
                    ),
                )

            set_llm_cache(GPTCache(init_gptcache))

        else:
            raise ValueError(f"Invalid LLM cache mode: {self.llm_cache_mode}")

    def add_langchain_api_route(
        self,
        url: str,
        langchain_object: Chain,
        streaming_mode: StreamingMode,
        methods: list[str] = ["POST"],
        **kwargs,
    ):
        """Adds a Langchain API route to the router.

        Args:
            url: The URL of the route.
            langchain_object: The Langchain object to use for the route.
            streaming_mode: The streaming mode to use for the route.
            methods: The HTTP methods to use for the route.
        """
        name_prefix = url.replace("/", "").title().replace("_", "")
        langchain_dependency = create_langchain_dependency(langchain_object)
        endpoint_request = create_request_from_langchain_dependency(
            langchain_dependency, name_prefix
        )
        response_model = (
            create_response_model_from_langchain_dependency(
                langchain_dependency, name_prefix
            )
            if streaming_mode == StreamingMode.OFF
            else None
        )

        endpoint = create_langchain_endpoint(
            endpoint_request,
            langchain_dependency,
            response_model,
            streaming_mode,
        )

        self.add_api_route(
            url,
            endpoint,
            response_model=response_model,
            methods=methods,
            **kwargs,
        )

        self.langchain_dependencies.append(langchain_dependency)

    def add_langchain_api_websocket_route(self, url: str, langchain_object: Chain):
        """Adds a Langchain API websocket route to the router.

        Args:
            url: The URL of the route.
            langchain_object: The Langchain object to use for the route.
        """
        langchain_dependency = create_langchain_dependency(langchain_object)
        endpoint = create_langchain_websocket_endpoint(
            WebSocket,
            langchain_dependency,
        )

        self.add_api_websocket_route(url, endpoint)

        self.langchain_dependencies.append(langchain_dependency)
